import inspect
from typing import Collection, List, Tuple
import numpy as np
from tqdm import tqdm
from sklearn.metrics import r2_score
import torch
from torch.nn import functional as F
from torch.optim.lr_scheduler import LRScheduler


def is_float(x) -> bool:
	if isinstance(x, (Collection, np.ndarray)):
		return is_float(x[0])
	return isinstance(x, (float, np.float_, np.float16, np.float16, np.float32, np.float64, np.float128, np.single, np.double))


def convert_to_tensor(x, dim_start=1) -> torch.Tensor:
	if 1 == dim_start:
		return torch.tensor(x, dtype=torch.float) if is_float(x[0]) else torch.tensor(x, dtype=torch.long)
	return torch.tensor(x, dtype=torch.float) if is_float(x[0][0]) else torch.tensor(x, dtype=torch.long)


def convert_data(X, y) -> Tuple[torch.Tensor, torch.Tensor]:
	if isinstance(X, (List, np.ndarray)):
		X = convert_to_tensor(X, 2)
	if isinstance(y, (List, np.ndarray)):
		y = convert_to_tensor(y)
	return X, y


def convert_data_acc(X, y) -> Tuple[torch.Tensor, torch.LongTensor]:
	if isinstance(X, (List, np.ndarray)):
		X = convert_to_tensor(X, 2)
	if isinstance(y, (List, np.ndarray)):
		y = torch.tensor(y, dtype=torch.long)
	return X, y


def convert_data_r2(X, y) -> Tuple[torch.Tensor, torch.FloatTensor]:
	if isinstance(X, (List, np.ndarray)):
		X = convert_to_tensor(X, 2)
	if isinstance(y, (List, np.ndarray)):
		y = torch.tensor(y, dtype=torch.float)
	return X, y


def cal_count(y) -> int:
	shape = y.shape
	return shape[0] if len(shape) == 1 else shape[0] * shape[1]


def evaluate(model, val_loader, device) -> float:
	total, steps = 0, 0
	total_loss = torch.Tensor([0.0]).to(device)
	model.eval()
	with torch.no_grad():
		if hasattr(model, 'fit'):
			for batch in val_loader:
				loss, logits = _fit(model, batch, device)
				total_loss += loss
				steps += 1
		else:
			for batch in val_loader:
				loss, logits = _train(model, batch, device)
				total_loss += loss
				steps += 1
	return total_loss.item() / steps


def r2_loss_logits(outputs, targets):
	if isinstance(outputs, tuple):
		loss, logits = outputs
	else:
		logits = outputs
		loss = F.mse_loss(outputs, targets.view(-1, 1))
	return loss, logits


def r2_evaluate(model, val_loader, device):
	total, steps = 0, 0
	total_loss = torch.Tensor([0.0]).to(device)
	labels, preds = [], []
	model.eval()
	with torch.no_grad():
		if hasattr(model, 'fit'):
			for batch in val_loader:
				outputs, y = _fit_y(model, batch, device)
				loss, logits = r2_loss_logits(outputs, y)
				total_loss += loss
				steps += 1
				labels.extend(y.detach().numpy())
				preds.extend(logits.detach().numpy().flatten())
		else:
			for batch in val_loader:
				outputs, y = _train_y(model, batch, device)
				loss, logits = r2_loss_logits(outputs, y)
				total_loss += loss
				steps += 1
				labels.extend(y.detach().tolist())
				preds.extend(logits.detach().numpy().flatten().tolist())
	return total_loss.item() / steps, r2_score(np.array(labels), np.array(preds))


def cal_correct(logits: torch.Tensor, y: torch.Tensor):
	if len(logits.size()) > 1:
		return (logits.argmax(-1) == y).sum()
	else:
		return (logits > 0.5).sum()


def acc_loss_logits(outputs, targets):
	if isinstance(outputs, tuple):
		loss, logits = outputs
	else:
		logits = outputs
		shapes = len(logits.size())
		if shapes == 2:
			loss = F.cross_entropy(outputs, targets)
		elif shapes == 3:
			loss = F.cross_entropy(outputs.reshape(-1, outputs.size(2)), targets.view(-1))
		else:
			raise ValueError("logits shapes must be 2 or 3")
	return loss, logits


def acc_evaluate(model, val_loader, device):
	total, steps = 0, 0
	total_loss, correct = [torch.Tensor([0.0]).to(device) for _ in range(2)]
	model.eval()
	with torch.no_grad():
		if hasattr(model, 'fit'):
			for batch in val_loader:
				outputs, y = _fit_y(model, batch, device)
				loss, logits = acc_loss_logits(outputs, y)
				total_loss += loss
				total += cal_count(y)
				correct += cal_correct(logits, y)
				steps += 1
		else:
			for batch in val_loader:
				outputs, y = _train_y(model, batch, device)
				loss, logits = acc_loss_logits(outputs, y)
				total_loss += loss
				total += cal_count(y)
				correct += cal_correct(logits, y)
				steps += 1
	return (total_loss.item() / steps), (correct.item() / total)


def _train(model, batch, device):
	if isinstance(batch, dict):
		batch = {key: value.to(device) if torch.is_tensor(value) else value for key, value in batch.items()}
		return model(**batch)
	else:
		batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
		return model(*batch)


def _fit(model, batch, device):
	if isinstance(batch, dict):
		batch = {key: value.to(device) if torch.is_tensor(value) else value for key, value in batch.items()}
		return model.fit(**batch)
	else:
		batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
		return model.fit(*batch)
	

def do_train(model, batch, optimizer, device):
	loss, logits = _train(model, batch, device)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss


def do_fit(model, batch, optimizer, device):
	loss, logits = _fit(model, batch, device)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss


def do_train_scheduler(model, batch, optimizer, device, scheduler: LRScheduler):
	loss, logits = _train(model, batch, device)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss


def do_fit_scheduler(model, batch, optimizer, device, scheduler: LRScheduler):
	loss, logits = _fit(model, batch, device)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss


def _train_y(model, batch, device):
	same_params = len(batch) == len(inspect.signature(model.forward).parameters)  # 判断所传参数个数是否与方法一致
	if isinstance(batch, dict):
		batch = {key: value.to(device) if torch.is_tensor(value) else value for key, value in batch.items()}
		if same_params:
			y = batch['labels'] if 'labels' in batch else batch['targets']
		else:  # 参数个数不一致，要把 'labels' 或 'targets' 从参数里剔除
			y = batch.pop('labels') if 'labels' in batch else batch.pop('targets')
		return model(**batch), y
	else:
		batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
		if same_params:
			return model(*batch), batch[-1]
		# 参数个数不一致，去掉最后一个
		return model(*(batch[:-1])), batch[-1]


def _fit_y(model, batch, device):
	if isinstance(batch, dict):
		batch = {key: value.to(device) if torch.is_tensor(value) else value for key, value in batch.items()}
		return model.fit(**batch), batch['labels'] if 'labels' in batch else batch['targets']
	else:
		batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
		return model.fit(*batch), batch[-1]


def do_train_acc(model, batch, optimizer, device):
	outputs, y = _train_y(model, batch, device)
	loss, logits = acc_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_fit_acc(model, batch, optimizer, device):
	outputs, y = _fit_y(model, batch, device)
	loss, logits = acc_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_train_scheduler_acc(model, batch, optimizer, device, scheduler: LRScheduler):
	outputs, y = _train_y(model, batch, device)
	loss, logits = acc_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_fit_scheduler_acc(model, batch, optimizer, device, scheduler: LRScheduler):
	outputs, y = _fit_y(model, batch, device)
	loss, logits = acc_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, cal_count(y), cal_correct(logits, y)


def train_epoch_base(model, train_loader, optimizer, device):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss = do_fit(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
	else:
		for batch in train_loader:
			loss = do_train(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
	
	return total_loss.item() / steps


def train_epoch_progress(model, train_loader, optimizer, device, epoch, epochs):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss = do_fit(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			loop.set_postfix(train_loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss = do_train(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			loop.set_postfix(train_loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	return total_loss.item() / steps


def train_epoch_scheduler(model, train_loader, optimizer, device, scheduler: LRScheduler):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss = do_fit_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
	else:
		for batch in train_loader:
			loss = do_train_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
	return total_loss.item() / steps


def train_epoch_scheduler_progress(model, train_loader, optimizer, device, scheduler, epoch, epochs):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss = do_fit_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			loop.set_postfix(train_loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss = do_train_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			loop.set_postfix(train_loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	return total_loss.item() / steps


def train_epoch(model, train_loader, optimizer, device, scheduler, epoch, epochs, show_progress):
	if show_progress:
		if scheduler is None:
			return train_epoch_progress(model, train_loader, optimizer, device, epoch, epochs)
		return train_epoch_scheduler_progress(model, train_loader, optimizer, device, scheduler, epoch, epochs)
	else:
		if scheduler is None:
			return train_epoch_base(model, train_loader, optimizer, device)
		return train_epoch_scheduler(model, train_loader, optimizer, device, scheduler)


def train_epoch_base_acc(model, train_loader, optimizer, device):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]).to(device) for _ in range(2)]
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, count, correct = do_fit_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	else:
		for batch in train_loader:
			loss, count, correct = do_train_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_progress_acc(model, train_loader, optimizer, device, epoch, epochs):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]).to(device) for _ in range(2)]
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, count, correct = do_fit_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(train_acc=f"{total_correct.item() / total:.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, count, correct = do_train_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(train_acc=f"{total_correct.item() / total:.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_scheduler_acc(model, train_loader, optimizer, device, scheduler: LRScheduler):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]).to(device) for _ in range(2)]
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, count, correct = do_fit_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	else:
		for batch in train_loader:
			loss, count, correct = do_train_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_scheduler_progress_acc(model, train_loader, optimizer, device, scheduler, epoch, epochs):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]).to(device) for _ in range(2)]
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, count, correct = do_fit_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(train_acc=f"{total_correct.item() / total:.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, count, correct = do_train_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(train_acc=f"{total_correct.item() / total:.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_acc(model, train_loader, optimizer, device, scheduler, epoch, epochs, show_progress):
	if show_progress:
		if scheduler is None:
			return train_epoch_progress_acc(model, train_loader, optimizer, device, epoch, epochs)
		return train_epoch_scheduler_progress_acc(model, train_loader, optimizer, device, scheduler, epoch, epochs)
	else:
		if scheduler is None:
			return train_epoch_base_acc(model, train_loader, optimizer, device)
		return train_epoch_scheduler_acc(model, train_loader, optimizer, device, scheduler)

################################################################################


def train_epoch_r2(model, train_loader, optimizer, device, scheduler, epoch, epochs, show_progress):
	if show_progress:
		if scheduler is None:
			return train_epoch_progress_r2(model, train_loader, optimizer, device, epoch, epochs)
		return train_epoch_scheduler_progress_r2(model, train_loader, optimizer, device, scheduler, epoch, epochs)
	else:
		if scheduler is None:
			return train_epoch_base_r2(model, train_loader, optimizer, device)
		return train_epoch_scheduler_r2(model, train_loader, optimizer, device, scheduler)


def train_epoch_base_r2(model, train_loader, optimizer, device):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	labels, preds = [], []
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, label, pred = do_fit_r2(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
	else:
		for batch in train_loader:
			loss, label, pred = do_train_r2(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
	return r2_score(np.array(labels), np.array(preds)), total_loss.item() / steps


def train_epoch_progress_r2(model, train_loader, optimizer, device, epoch, epochs):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	labels, preds = [], []
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, label, pred = do_fit_r2(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
			loop.set_postfix(train_r2=f"{r2_score(np.array(labels), np.array(preds)):.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, label, pred = do_train_r2(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
			loop.set_postfix(train_r2=f"{r2_score(np.array(labels), np.array(preds)):.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return r2_score(np.array(labels), np.array(preds)), total_loss.item() / steps


def train_epoch_scheduler_r2(model, train_loader, optimizer, device, scheduler: LRScheduler):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	labels, preds = [], []
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, label, pred = do_fit_scheduler_r2(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
	else:
		for batch in train_loader:
			loss, label, pred = do_train_scheduler_r2(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
	return r2_score(np.array(labels), np.array(preds)), total_loss.item() / steps


def train_epoch_scheduler_progress_r2(model, train_loader, optimizer, device, scheduler, epoch, epochs):
	steps = 0
	total_loss = torch.Tensor([0.0]).to(device)
	labels, preds = [], []
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{epochs}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, label, pred = do_fit_scheduler_r2(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
			loop.set_postfix(train_r2=f"{r2_score(np.array(labels), np.array(preds)):.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, label, pred = do_train_scheduler_r2(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			labels.extend(label)
			preds.extend(pred)
			loop.set_postfix(train_r2=f"{r2_score(np.array(labels), np.array(preds)):.4f}",
			                 train_loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return r2_score(np.array(labels), np.array(preds)), total_loss.item() / steps


def do_train_r2(model, batch, optimizer, device):
	outputs, y = _train_y(model, batch, device)
	loss, logits = r2_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, y.detach().tolist(), logits.detach().numpy().flatten().tolist()


def do_fit_r2(model, batch, optimizer, device):
	outputs, y = _fit_y(model, batch, device)
	loss, logits = r2_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, y.detach().tolist(), logits.detach().numpy().flatten().tolist()


def do_train_scheduler_r2(model, batch, optimizer, device, scheduler: LRScheduler):
	outputs, y = _train_y(model, batch, device)
	loss, logits = r2_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, y.detach().tolist(), logits.detach().numpy().flatten().tolist()


def do_fit_scheduler_r2(model, batch, optimizer, device, scheduler: LRScheduler):
	outputs, y = _fit_y(model, batch, device)
	loss, logits = r2_loss_logits(outputs, y)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, y.detach().tolist(), logits.detach().numpy().flatten().tolist()



import math
import torch
from torch import nn
from typing import Optional
import torch.nn.functional as F
from .layer import CNNLayer, RNNLayer
from .embedding import RotaryEmbedding, SelfRotaryEmbedding


class RotaryAttention(nn.Module):
	def __init__(self, embed_dim: int, num_heads: int = 1, max_length: int = 100000, theta: float = 10000.0,
	             dropout: float = 0., bias: bool = True):
		"""
		:param embed_dim: word embedding维度
		:param num_heads: Number of parallel attention heads.
		:param max_length: max length of sequences
		:param theta (float, optional): Scaling factor for frequency computation. Defaults to 10000.0.
		"""
		super().__init__()
		self.rotary_embedding = RotaryEmbedding(embed_dim, max_length, theta)
		self.attention = nn.MultiheadAttention(embed_dim, num_heads, dropout, bias)
	
	def forward(self, inputs: torch.Tensor, mask: Optional[torch.Tensor] = None, start_pos: int = 0):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:param mask: [(batch_size, sequence_length, embed_dim)]
		:param start_pos:
		:return: [(batch_size, sequence_length, embed_dim)]
		"""
		query, key = self.rotary_embedding(inputs, start_pos)
		return self.attention(query, key, inputs, attn_mask=mask)


def attention(query: torch.Tensor, key: torch.Tensor, value: torch.Tensor):
	batch_size, length, d_k = key.size()
	key = key.contiguous()
	scores = torch.matmul(query, key.view(batch_size, d_k, length)) / math.sqrt(d_k)
	scores = F.softmax(scores, dim=-1)
	return torch.matmul(scores, value).sum(dim=1)


class ClassSelfAttention(nn.Module):

	def __init__(self, embed_dim: int):
		super().__init__()
		self.w_omega = nn.Parameter(torch.empty(embed_dim, embed_dim))
		self.u_omega = nn.Parameter(torch.empty(embed_dim, 1))
		# self.w_omega = nn.Parameter(torch.Tensor(embed_dim, embed_dim))
		# self.u_omega = nn.Parameter(torch.Tensor(embed_dim, 1))
		nn.init.uniform_(self.w_omega, -0.1, 0.1)
		nn.init.uniform_(self.u_omega, -0.1, 0.1)

	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, embed_dim)]
		"""
		u = torch.tanh(torch.matmul(inputs, self.w_omega))
		att = torch.matmul(u, self.u_omega)
		att_score = F.softmax(att, dim=1)
		value = inputs * att_score
		return torch.sum(value, dim=1)


class RotaryClassSelfAttention(nn.Module):

	def __init__(self, embed_dim: int, max_length: int = 100000, theta: float = 10000.0):
		super().__init__()
		self.rotary_embedding = SelfRotaryEmbedding(embed_dim, max_length, theta)
		self.w_omega = nn.Parameter(torch.Tensor(embed_dim, embed_dim))
		self.u_omega = nn.Parameter(torch.Tensor(embed_dim, 1))
		nn.init.uniform_(self.w_omega, -0.1, 0.1)
		nn.init.uniform_(self.u_omega, -0.1, 0.1)

	def forward(self, inputs: torch.Tensor, start_pos: int = 0):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:param start_pos:
		:return: [(batch_size, embed_dim)]
		"""
		rotary = self.rotary_embedding(inputs, start_pos)
		u = torch.tanh(torch.matmul(rotary, self.w_omega))
		att = torch.matmul(u, self.u_omega)
		att_score = F.softmax(att, dim=1)
		value = inputs * att_score
		return torch.sum(value, dim=1)


class MultiHeadClassSelfAttention(nn.Module):

	def __init__(self, embed_dim: int, num_heads: int = 1):
		super().__init__()
		self.num_heads = num_heads
		self.attention = ClassSelfAttention(num_heads * embed_dim)

	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, num_heads * embed_dim)]
		"""
		if self.num_heads > 1:
			inputs = torch.concat([inputs for _ in range(self.num_heads)], dim=-1)
		return self.attention(inputs)


class RNNAttention(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model import RNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNAttention(embed_dim, out_features=len(classes))
	"""

	def __init__(self, embed_dim: int, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1,
				 out_features: int = 2, rnn=nn.GRU, bidirectional=True, layer_norm=False, residual=False,
				 dropout: float = 0.0):
		"""
		如果是英文、分词, 不用residual 效果比较好
		
		:param embed_dim: RNN的input_size，word embedding维度
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param residual: 是否残差
		:param dropout：
		"""
	
		super().__init__()
		self.residual = residual
		rnn_output_size = hidden_size << 1 if bidirectional else hidden_size
		attn_embed_dim = rnn_output_size + embed_dim if residual else rnn_output_size
		self.rnn = RNNLayer(embed_dim, hidden_size, num_layers, rnn, bidirectional, layer_norm, dropout)
		self.attention = MultiHeadClassSelfAttention(attn_embed_dim, num_heads)
	
		# 定义全连接层
		self.fc = nn.Linear(attn_embed_dim * num_heads, out_features)
		if 0.0 < dropout < 1.0:
			self.fc = nn.Sequential(
				nn.Dropout(dropout),
				self.fc
			)

	def forward(self, inputs: torch.Tensor, sequence_lengths: Optional[torch.IntTensor] = None):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:param sequence_lengths: [torch.IntTensor] 序列实际长度
		:return: [(batch_size, out_features)]
		"""
		output = self.rnn(inputs, sequence_lengths)  # [(batch_size, sequence_length, rnn_output_size)]
		if self.residual:
			output = torch.cat((output, inputs), dim=2)
		output = self.attention(output)  # [(batch_size, attn_embed_dim)]
		return self.fc(output)
	
	def fit(self, inputs: torch.Tensor, sequence_lengths: Optional[torch.IntTensor] = None, labels: torch.LongTensor = None):
		logits = self.forward(inputs, sequence_lengths)
		return F.cross_entropy(logits, labels), logits


class CNNRNNAttention(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model import CNNRNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = CNNRNNAttention(embed_dim, out_features=len(classes), vocab_size=tokenizer.vocab_size)
	"""
	
	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = 64, kernel_sizes=(2, 3, 4),
	             activation=None, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1,
	             out_features: int = 2, rnn=nn.GRU, bidirectional=True, layer_norm=False, batch_norm=False,
	             residual=False, dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: RNN的input_size，word embedding维度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param out_features:
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param residual: 是否残差
		:param batch_norm: 是否层批则化
		:param dropout：
		:param bias：
		"""
		
		super().__init__()
		cnn_channels = cnn_channels or embed_dim
		self.cnn = CNNLayer(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, layer_norm, batch_norm, bias)
		self.attn = RNNAttention(cnn_channels, hidden_size, num_layers, num_heads, out_features, rnn, bidirectional,
		                         layer_norm, residual, dropout)
		
	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, out_features)]
		"""
		output = self.cnn(inputs)
		return self.attn(output)
	
	def fit(self, inputs: torch.Tensor, labels: torch.LongTensor):
		logits = self.forward(inputs)
		return F.cross_entropy(logits, labels), logits


class RNNCNNAttention(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model import CNNRNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNCNNAttention(embed_dim, out_features=len(classes), vocab_size=tokenizer.vocab_size)
	"""
	
	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = 64, kernel_sizes=(2, 3, 4),
	             activation=None, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1,
	             out_features: int = 2, rnn=nn.GRU, bidirectional=True, layer_norm=False, batch_norm=False,
	             dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: RNN的input_size，word embedding维度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param out_features:
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param dropout：
		:param bias：
		"""
		super().__init__()
		cnn_channels = cnn_channels or embed_dim
		rnn_output_size = (hidden_size << 1) if bidirectional else hidden_size
		self.rnn = RNNLayer(embed_dim, hidden_size, num_layers, rnn, bidirectional, layer_norm, dropout)
		self.cnn = CNNLayer(rnn_output_size, seq_length, cnn_channels, kernel_sizes, activation, layer_norm, batch_norm, bias)
		self.attention = MultiHeadClassSelfAttention(cnn_channels, num_heads)
	
		# 定义全连接层
		self.fc = nn.Linear(cnn_channels * num_heads, out_features)
		if 0.0 < dropout < 1.0:
			self.fc = nn.Sequential(
				nn.Dropout(dropout),
				self.fc
			)

	def forward(self, inputs: torch.Tensor, sequence_lengths: torch.IntTensor = None):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:param sequence_lengths: [torch.IntTensor] 序列实际长度
		:return: [(batch_size, out_features)]
		"""
		output = self.rnn(inputs, sequence_lengths)
		output = self.cnn(output)
		output = self.attention(output)
		return self.fc(output)
	
	def fit(self, inputs: torch.Tensor, sequence_lengths: Optional[torch.IntTensor] = None, labels: torch.LongTensor = None):
		logits = self.forward(inputs, sequence_lengths)
		return F.cross_entropy(logits, labels), logits


class ResRNNCNNAttention(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model import CNNRNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = ResRNNCNNAttention(embed_dim, out_features=len(classes), vocab_size=tokenizer.vocab_size)
	"""
	
	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = 64, kernel_sizes=(2, 3, 4),
	             activation=None, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1,
	             out_features: int = 2, rnn=nn.GRU, bidirectional=True, layer_norm=False, batch_norm=False,
	             dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: RNN的input_size，word embedding维度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param out_features:
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param dropout：
		:param bias：
		"""
		super().__init__()
		cnn_channels = cnn_channels or embed_dim
		rnn_output_size = (hidden_size << 1) if bidirectional else hidden_size
		self.rnn = RNNLayer(embed_dim, hidden_size, num_layers, rnn, bidirectional, layer_norm, dropout)
		self.cnn = CNNLayer(rnn_output_size, seq_length, cnn_channels, kernel_sizes, activation, layer_norm, batch_norm, bias)
		self.attention = MultiHeadClassSelfAttention(cnn_channels, num_heads)
		self.res_attention = MultiHeadClassSelfAttention(embed_dim + rnn_output_size, num_heads)
	
		# 定义全连接层
		self.fc = nn.Linear((embed_dim + cnn_channels + rnn_output_size) * num_heads, out_features)
		if 0.0 < dropout < 1.0:
			self.fc = nn.Sequential(
				nn.Dropout(dropout),
				self.fc
			)
		# if batch_norm:
		# 	self.fc = nn.Sequential(
		# 		nn.BatchNorm1d((embed_dim + cnn_channels + rnn_output_hidden_size) * num_heads),
		# 		self.fc
		# 	)

	def forward(self, inputs: torch.Tensor, sequence_lengths: torch.IntTensor = None):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:param sequence_lengths: [torch.IntTensor] 序列实际长度
		:return: [(batch_size, out_features)]
		"""
		rnn_output = self.rnn(inputs, sequence_lengths)  # [(batch_size, sequence_length, 2 * hidden_size)]
		rnn_cnn_output = self.cnn(rnn_output)     # [(batch_size, seq_length * len(kernel_sizes), cnn_channels)]
		rnn_cnn_attn_output = self.attention(rnn_cnn_output)   # [(batch_size, num_heads * out_channels)]
		
		# [(batch_size, num_heads * (embed_dim + 2 * hidden_size))]
		res_output = self.res_attention(torch.cat((inputs, rnn_output), dim=2))
		
		output = torch.cat([res_output, rnn_cnn_attn_output], dim=1)
		return self.fc(output)
	
	def fit(self, inputs: torch.Tensor, sequence_lengths: Optional[torch.IntTensor] = None, labels: torch.LongTensor = None):
		logits = self.forward(inputs, sequence_lengths)
		return F.cross_entropy(logits, labels), logits



import os
import re
import operator
import numpy as np
from pathlib import Path
from functools import reduce
from itertools import tee, chain
from typing import List, Union, Iterable, Collection, Tuple
from concurrent.futures import as_completed, ProcessPoolExecutor

UTF8 = 'utf-8'
REMOVE_CHARS = '[·’!"\#$%&\'()＃！（）*+,-./:;<=>?\@，。：?￥★、…．＞【】［］《》？“”‘’\[\\]^_`{|}~]+'
REMOVE_CHARS_PATTERN = re.compile(REMOVE_CHARS)
EN_PATTERN = re.compile('[^A-Za-z]+')
EN_CHAR_PATTERN = re.compile(r'([\W])')
CN_CHAR_PATTERN = re.compile(r'([\u4e00-\u9fa5])')


def convert_labels(labels) -> Tuple[np.ndarray, list]:
	"""
	:return (labels, classes)
	"""
	labels = labels.astype('category').cat
	return labels.codes.to_numpy(), labels.categories.tolist()


def get_text_length(text: str, lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None) -> int:
	if cut_fn is None and lang == 'cn' and cut_type == 'word':
		import jieba
		import logging
		jieba.setLogLevel(logging.INFO)
		cut_fn = jieba.lcut
	cuts = cut(text, lang, cut_type, keep_punctuation, cut_fn=cut_fn)
	return len(cuts)


def get_texts_max_length(texts: List[str], lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None) -> int:
	if cut_fn is None and lang == 'cn' and cut_type == 'word':
		import jieba
		import logging
		jieba.setLogLevel(logging.INFO)
		cut_fn = jieba.lcut
	batch_cuts = batch_cut(texts, lang, cut_type, keep_punctuation, cut_fn=cut_fn)
	return max(map(lambda s: len(s), batch_cuts))


def get_df_text_labels(df, text_col: str, label_col: str, min_length: int = None, max_length: int = None):
	"""Returns (texts, labels, classes)"""
	
	if min_length or max_length:
		df['length'] = df[text_col].apply(lambda x: get_text_length(x))
	if min_length:
		df = df[df['length'] >= min_length]
	if max_length:
		df = df[df['length'] <= max_length]
	
	_labels, _classes = convert_labels(df[label_col])
	_texts = df[text_col].values
	return _texts, _labels, _classes


# -----------------------------------------------------------Read file-------------------------------------------------
def read_file(path: Union[str, Path], encoding=UTF8):
	"""
    读取文件内容
    :param path: 文件名，不能是文件夹
    :param encoding: 编码
    :return: 包含非空文本行的生成器，如 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
    """
	# if Path(path).stat().st_size <= 10000: #1048576000: # 小于等于1G
	with open(path, encoding=encoding) as f:
		lines = f.readlines()
	return [line.strip() for line in lines if line.strip()]


def read_large_file(path: Union[str, Path], encoding=UTF8):
	with open(path, encoding=encoding) as f:
		for line in f:
			line = line.strip()
			if line:
				yield line


async def async_read_file(path: Union[str, Path], encoding=UTF8):
	return read_file(path, encoding)


def read_corpus_files(path: Union[str, Path], encoding=UTF8, pattern='*', func=read_file, async_func=async_read_file):
	"""
    读取文件或文件夹下所有符合条件的文件
    :param path: 文件或文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
    :param encoding: 编码
    :param pattern: 当path是文件夹的时候，会根据此后缀过滤文件
    :param func: 具体读取文件的读函数，默认是read_file，可替换。注意：其函数签名为 function_name(path: str, encoding: str) -> corpus: Iterable[str]
    :param async_func: 传入文件夹时的读函数，用协程读取每个文件，默认是async_read_file，可替换。注意：其函数签名为 async function_name(path: str, encoding: str) -> corpus: Iterable[str]
    :return: 非空文本行，如 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
    """
	if isinstance(path, str):
		path = Path(path)
	
	if path.is_file():
		return func(path, encoding)
	
	import asyncio
	loop = asyncio.get_event_loop()
	tasks = [loop.create_task(async_func(file, encoding)) for file in path.rglob(pattern)]  # 这里不能用map，否则读不出数据
	wait_coro = asyncio.wait(tasks)
	loop.run_until_complete(wait_coro)
	all_lines = (task.result() for task in tasks)
	loop.close()
	
	return reduce(operator.iconcat, all_lines, [])


def load_embedding(path: str, is_large_file=False):
	if path.endswith('.bz2'):
		return load_embedding_from_bz2(path, is_large_file)
	return load_embedding_nomal(path, is_large_file)


def load_embedding_from_bz2(path: str, is_large_file=False):
	import bz2
	if is_large_file:
		with bz2.open(path, 'r') as f:
			tokens, vecs = _get_token_vecs(f, require_decode_token=True)
		return list(tokens), list(vecs)
	
	with bz2.open(path, 'r') as f:
		lines = f.readlines()
	return _handle_lines(lines, require_decode_token=True)


def load_embedding_nomal(path: str, is_large_file=False):
	if is_large_file:
		with open(path, 'r', encoding=UTF8) as f:
			tokens, vecs = _get_token_vecs(f, require_decode_token=False)
		return list(tokens), list(vecs)
	
	with open(path, 'r', encoding=UTF8) as f:
		lines = f.readlines()
	return _handle_lines(lines, require_decode_token=False)


def _get_token_vecs(f, require_decode_token):
	token_vec = (_handle_line(line, require_decode_token) for line in f if len(line.rstrip().split()) > 2)
	return zip(*token_vec)


def _handle_lines(lines: Iterable[str], require_decode_token: bool, batch_size=8192):
	if len(lines[0].split()) <= 2:
		lines = lines[1:]
	
	length = len(lines)
	if length <= (batch_size << 2):
		token_vecs = list(map(lambda line: _handle_line(line, require_decode_token), lines))
		tokens, vecs = zip(*token_vecs)
		return list(tokens), list(vecs)
	
	splits = [lines[i:i + batch_size] for i in range(0, length, batch_size)]
	n_jobs = min(len(splits), os.cpu_count() - 1)
	with ProcessPoolExecutor(n_jobs) as executor:
		results = [executor.submit(_order_handle_lines, i, split, require_decode_token) for i, split in enumerate(splits)]
		results = [r.result() for r in as_completed(results)]
	results.sort(key=lambda x: x[0], reverse=False)
	results = [r[1] for r in results]
	token_vecs = list(chain(*results))  # [[str,],] => [str,]
	tokens, vecs = zip(*token_vecs)
	return list(tokens), list(vecs)


def _order_handle_lines(order: int, lines: Iterable[str], require_decode_token: bool):
	return order, list(map(lambda line: _handle_line(line, require_decode_token), lines))


def _handle_line(line: str, require_decode_token: bool):
	def get_vec(elems):
		return list(map(float, elems))
	
	elems = line.rstrip().split()
	return elems[0].decode(UTF8) if require_decode_token else elems[0], get_vec(elems[1:])


# --------------------------------------------------Token cut----------------------------------------------------------
def cut_char(sentence: str):
	"""
    把句子按字分开，不破坏英文结构
    把所有的中文分开，同时希望英文和数字不能被拆分
    """
	parts = EN_CHAR_PATTERN.split(sentence)
	parts = filter(str.strip, parts)
	results = map(_cn_char_handle, parts)
	return list(chain(*results))


def batch_cut(text: Collection[str], lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None):
	"""
    多句话批量分词
    :param text: 多句话，即多行
    :param lang: 哪国语言，支持cn和en
    :param cut_type: 按词还是字分，支持word和char
    :param keep_punctuation: 是否保留标点符号
    :param cut_fn
    :return: 分词后的list(2维)
    """
	if lang == 'cn':
		replace_char = ''
		if cut_type == 'word':
			if keep_punctuation:
				def fn(s):
					return cut_fn(s.strip())
			else:
				def fn(s):
					return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, s.strip()))
		else:
			if cut_fn is None:
				cut_fn = cut_char
			
			if keep_punctuation:
				def fn(s):
					return cut_fn(s.strip())
			else:
				def fn(s):
					return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, s.strip()))
		return map(fn, text)
		# return [cut_char(re.sub(REMOVE_CHARS_PATTERN, '', line.strip())) for line in text]
	
	if lang == 'en':
		replace_char = ' '
		if cut_type == 'word':
			def fn(s):
				return re.sub(REMOVE_CHARS_PATTERN, replace_char, s).strip().lower().split()
		else:
			if keep_punctuation:
				def fn(s):
					return list(s.strip().lower())
			else:
				def fn(s):
					return list(re.sub(EN_PATTERN, replace_char, s).strip().lower())
		return map(fn, text)
		# return [fn(line) for line in text]
	
	raise NotImplementedError(f'暂时未实现"{lang}"的分词功能')


def cut(sentence: str, lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None):
	"""
    单句话分词
    :param sentence: 单句话，即一行
    :param lang: 哪国语言，支持cn和en
    :param cut_type: 按词还是字分，支持word和char
    :param keep_punctuation: 是否保留标点符号
    :param cut_fn
    :return: 分词后的list(1维)
    """
	import re
	if lang == 'cn':
		replace_char = ''
		if cut_type == 'word':
			if keep_punctuation:
				return list(cut_fn(sentence.strip()))
			else:
				return list(cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence.strip())))
		else:
			if cut_fn is None:
				cut_fn = cut_char
			
			if keep_punctuation:
				return cut_fn(sentence.strip())
			else:
				return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence.strip()))
	
	if lang == 'en':
		replace_char = ' '
		if cut_type == 'word':
			return re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence).strip().lower().split()
		else:
			if keep_punctuation:
				return list(sentence.strip().lower())
			else:
				return list(re.sub(EN_PATTERN, replace_char, sentence).strip().lower())
	
	raise NotImplementedError(f'暂时未实现"{lang}"的分词功能')


def filter_stop_words(cut_corpus: Iterable[List[str]], stop_words: Collection[str]):
	return map(lambda line: [word for word in line if word not in stop_words], cut_corpus)
	# return map(lambda line: list(filter(lambda word: word not in stop_words, line)), cut_corpus)


# -------------------------------------------------pad function--------------------------------------------------------
def pad(ids: List[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
        bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	"""
    pad token list(1维)
    :return: padded list, valid_size size
    """
	assert padding_side in ('right', 'left'), '参数padding_side只能是"right"或"left".'
	size = len(ids)
	if bos and eos:
		size += 2
		if max_length and truncation and size > max_length:
			return [bos_id] + ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return [bos_id] + ids + [eos_id] + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + [bos_id] + ids + [eos_id], size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return [bos_id] + ids + [eos_id], size
	elif bos:
		size += 1
		if max_length and truncation and size > max_length:
			return [bos_id] + ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return [bos_id] + ids + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + [bos_id] + ids, size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return [bos_id] + ids, size
	elif eos:
		size += 1
		if max_length and truncation and size > max_length:
			return ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return ids + [eos_id] + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + ids + [eos_id], size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return ids + [eos_id], size
	else:
		if max_length and truncation and size > max_length:
			return ids[:max_length], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return ids + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + ids, size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return ids, size


def batch_pad(batch: Union[map, Iterable[Iterable[int]]], max_length: int = None, truncation=True,
              padding=True, padding_side='right', bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	"""
    :param batch: (2维)
    :param max_length:
    :param truncation:
    :param padding_side:
    :param bos:
    :param eos:
    :param pad_id:
    :param bos_id:
    :param eos_id:
    :return:
    """
	if not max_length and (truncation or padding):
		batch, batch_copy = tee(batch, 2)
		max_length = max(map(len, batch_copy))
		if bos:
			max_length += 1
		if eos:
			max_length += 1
	ids, sizes = zip(*map(lambda ids: pad(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id),batch))
	return list(ids), list(sizes)


def _pad_mask(ids: Iterable[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
              bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	input_ids, real_size = pad(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id)
	size = len(input_ids)
	if size > real_size:
		if padding_side == 'right':
			return input_ids, [1] * real_size + [0] * (size - real_size)
		return input_ids, [0] * (size - real_size) + [1] * real_size
	else:
		return input_ids, [1] * size


def pad_mask(ids: Iterable[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
             bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	input_ids, mask_ids = _pad_mask(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id,
	                                eos_id)
	return {'input_ids': input_ids, 'mask_ids': mask_ids}


def batch_pad_mask(batch: Union[map, Iterable[Iterable[int]]], max_length: int = None, truncation=True,
                   padding=True, padding_side='right', bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	if not max_length and (truncation or padding):
		batch, batch_copy = tee(batch, 2)
		max_length = max(map(len, batch_copy))
		if bos:
			max_length += 1
		if eos:
			max_length += 1
	ids = list(
		map(lambda ids: _pad_mask(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id),
		    batch))
	input_ids, mask_ids = zip(*ids)
	return {'input_ids': list(input_ids), 'mask_ids': list(mask_ids)}


def _cn_char_handle(part):
	chars = CN_CHAR_PATTERN.split(part)
	return filter(str.strip, chars)



from logging import basicConfig, INFO, getLogger

logger = getLogger(__name__)
basicConfig(level=INFO, format='[%(asctime)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


def info(message):
	logger.info(message)



from ._dataset import ListDataset, TokenDataset, SameLengthTokenDataset, TextDataset, TextDFDataset
from ._collector import TextVecCollator, TokenizeCollator, PaddingTokenCollator, PaddingLongTensorCollector

__all__ = [
	"ListDataset",
	"TokenDataset",
	"SameLengthTokenDataset",
	"TextDataset",
	"TextDFDataset",
	"TextVecCollator",
	"TokenizeCollator",
	"PaddingTokenCollator",
	"PaddingLongTensorCollector",
]



from ._tokenize_vec import TokenizeVec, BertTokenizeVec, ErnieTokenizeVec, AlbertTokenizeVec
from ._utils import train_test_set

__all__ = [
	"TokenizeVec",
	"BertTokenizeVec",
	"ErnieTokenizeVec",
	"AlbertTokenizeVec",
	"train_test_set"
]



import torch
import numpy as np
from typing import Union, List, Tuple

from nlpx.llm import TokenizeVec
from nlpx.text_token import BaseTokenizer, TokenEmbedding
from nlpx.text_token.utils import get_texts_max_length


class TextVecCollator:
	
	def __init__(self, tokenize_vec: Union[TokenizeVec, TokenEmbedding], max_length: int = None, **kwargs):
		self.tokenize_vec = tokenize_vec
		self.max_length = max_length
		self.kwargs = kwargs
	
	def __call__(self, examples):
		texts, labels = zip(*examples)
		labels = torch.LongTensor(np.array(labels))
		
		if isinstance(self.tokenize_vec, TokenizeVec):
			max_length = get_texts_max_length(texts, cut_type='char') + 2
			max_length = min(max_length, self.max_length) if self.max_length and self.max_length > 0 else max_length
			return self.tokenize_vec.encode_plus(texts, max_length=max_length, padding='max_length',
			                                     truncation=True, add_special_tokens=True,
			                                     return_token_type_ids=True, return_attention_mask=True,
			                                     return_tensors='pt', **self.kwargs), labels
		elif isinstance(self.tokenize_vec, TokenEmbedding):
			max_length = get_texts_max_length(texts, cut_type=self.tokenize_vec.cut_type,
			                                  lang=self.tokenize_vec.lang, cut_fn=self.tokenize_vec.cut_fn)
			max_length = min(max_length, self.max_length) if self.max_length and self.max_length > 0 else max_length
			return self.tokenize_vec(texts, max_length, **self.kwargs), labels
		
		raise ValueError("Invalid tokenize_vec, it must be a TokenizeVec or TokenEmbedding.")


class TokenizeCollator:
	
	def __init__(self, tokenizer, max_length: int = None, **kwargs):
		self.tokenizer = tokenizer
		self.max_length = max_length
		self.kwargs = kwargs
	
	def __call__(self, examples):
		texts, labels = zip(*examples)
		labels = torch.LongTensor(np.array(labels))
		
		if isinstance(self.tokenizer, BaseTokenizer):
			max_length = get_texts_max_length(texts, cut_type=self.tokenizer.cut_type, lang=self.tokenizer.lang,
			                                  cut_fn=self.tokenizer.cut_fn)
			max_length = min(max_length, self.max_length) if self.max_length else max_length
			return torch.LongTensor(self.tokenizer.batch_encode(texts, max_length, **self.kwargs)), labels
		
		max_length = get_texts_max_length(texts, cut_type='char') + 2
		max_length = min(max_length, self.max_length) if self.max_length and self.max_length > 0 else max_length
		result = self.tokenizer.batch_encode_plus(texts, max_length=max_length, padding='max_length',
		                                          return_token_type_ids=True, return_attention_mask=True,
		                                          truncation=True, add_special_tokens=True, return_tensors='pt',
		                                          **self.kwargs)
		result['labels'] = labels
		return result


class PaddingTokenCollator:
	"""与TokenDataset配合使用, 只有token, label两列数数

	Examples
	--------
	>>> tokenizer = PaddingTokenizer(texts=texts)
	>>> X_train = tokenizer.batch_encode(texts, padding=False)
	>>> train_set = TokenDataset(X_train, y_train)
	>>> model_wrapper = ClassModelWrapper(classes=classes)
	>>> model_wrapper.train(model, train_set, early_stopping_rounds=5, show_progress=False,
	>>>                     collate_fn=PaddingTokenCollator(tokenizer.pad, return_sequence_length=True))
	"""
	
	def __init__(self, pad_func, max_length: int = None, truncation=True, padding_side='right',
	             return_sequence_length=False, bos=False, eos=False):
		self.pad_func = pad_func
		self.max_length = max_length
		self.truncation = truncation
		self.padding_side = padding_side
		self.return_sequence_length = return_sequence_length
		self.bos, self.eos = bos, eos
	
	def __call__(self, examples):
		tokens, labels = zip(*examples)
		labels = torch.LongTensor(np.array(labels))
		
		max_length = max(map(lambda x: len(x), tokens))
		max_length = min(max_length, self.max_length) if self.max_length and self.max_length > 0 else max_length
		params = {'truncation': self.truncation, 'padding_side': self.padding_side}
		if self.bos:
			params['bos'] = self.bos
		if self.eos:
			params['eos'] = self.eos
		
		if self.return_sequence_length:
			params['return_sequence_length'] = self.return_sequence_length
			ids, sequence_lengths = self.pad_func(tokens, max_length, **params)
			return torch.LongTensor(ids), torch.IntTensor(sequence_lengths), labels
		
		ids = self.pad_func(tokens, max_length, **params)
		if isinstance(ids, Tuple):
			ids = ids[0]
		return torch.LongTensor(ids), labels


class PaddingLongTensorCollector:
	""" 可以有多列数据 """
	
	def __init__(self, pad_func):
		self.pad_func = pad_func
	
	def __call__(self, batch):
		batch = (self.pad_func(x) if isinstance(x[0], List) else x for x in zip(*batch))
		return tuple(torch.tensor(X, dtype=torch.long) for x in batch)
	


from ._text_cnn import TextCNN
from ._attention import RotaryAttention, attention, ClassSelfAttention, MultiHeadClassSelfAttention, \
	RotaryClassSelfAttention, RNNAttention, CNNRNNAttention, RNNCNNAttention, ResRNNCNNAttention


__all__ = [
	"TextCNN",
	"RotaryAttention",
	"attention",
	"ClassSelfAttention",
	"MultiHeadClassSelfAttention",
	"RotaryClassSelfAttention",
	"RNNAttention",
	"CNNRNNAttention",
	"RNNCNNAttention"
]



import os
from setuptools import setup


# INSTALL_PACKAGES = open(path.join(DIR, 'requirements.txt')).read().splitlines()
def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), 'r', encoding='UTF-8') as fp:
        return fp.read()


long_description = read("README.rst")


setup(
    name='NLPX',
    packages=['nlpx', 'nlpx.model', 'nlpx.llm', 'nlpx.text_token', 'nlpx.dataset', 'nlpx.training'],
    description="A tool set for NLP.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.6.9',
    url='https://gitee.com/summry/nlpx',
    author='summy',
    author_email='xiazhongbiao@126.com',
    keywords=['NLP', 'nlp', 'AI', 'llm', 'GPT', 'Machine learning', 'Deep learning', 'tokenize', 'torch'],
    package_data={
        # include json and txt files
        '': ['*.rst', '*.dtd', '*.tpl'],
    },
    include_package_data=True,
    python_requires='>=3.6',
    zip_safe=False
)




import torch
from torch import nn
import torch.nn.functional as F
from typing import Union, List, Optional
from ._text_cnn import TextCNN
from ._attention import RNNAttention, CNNRNNAttention, RNNCNNAttention, ResRNNCNNAttention, RotaryClassSelfAttention
from .embedding import EmbeddingLayer, CNNEmbedding
from nlpx.norm import RMSNorm


class EmbeddingClassifier(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import TokenEmbedding
	>>> from nlpx.model import TextCNN, RNNAttention
	>>> from nlpx.model.classifier import EmbeddingClassifier
	>>> tokenizer = TokenEmbedding(pretrained_path)
	>>> attn = RNNAttention(tokenizer.embed_dim, num_heads=2, out_features=len(classes))
	>>> classifier = EmbeddingClassifier(atten, embedding=tokenizer.embedding)
	>>> classifier = EmbeddingClassifier(atten, vocab_size=tokenizer.vocab_size, embed_dim=tokenizer.embed_dim)
	"""

	def __init__(self, classifier, embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
	             embed_dim: int = None, padding_idx: Optional[int] = None):
		"""
		:param classifier: 分类器
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param vocab_size: vocab size
		:param embed_dim: word embedding维度
		:param padding_idx
		"""
		super().__init__()
		self.classifier = classifier
		self.embedding = EmbeddingLayer(embed_dim, vocab_size, embedding, padding_idx)

	def forward(self, input_ids: torch.Tensor):
		embedding = self.embedding(input_ids)
		return self.classifier(embedding)
	
	def fit(self, input_ids: torch.Tensor, labels: torch.LongTensor):
		logits = self.forward(input_ids)
		return F.cross_entropy(logits, labels), logits
	

class MaskEmbeddingClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import TokenEmbedding
	>>> from nlpx.model import TextCNN, RNNAttention
	>>> from nlpx.model.classifier import EmbeddingClassifier
	>>> tokenizer = TokenEmbedding(pretrained_path)
	>>> attn = RNNAttention(tokenizer.embed_dim, num_heads=2, out_features=len(classes))
	>>> classifier = MaskEmbeddingClassifier(atten, embedding=tokenizer.embedding)
	>>> classifier = MaskEmbeddingClassifier(atten, vocab_size=tokenizer.vocab_size, embed_dim=tokenizer.embed_dim)
	"""
	
	def forward(self, input_ids: torch.Tensor, mask: Optional[torch.Tensor] = None):
		embedding = self.embedding(input_ids)
		return self.classifier(embedding, mask)
	
	def fit(self, input_ids: torch.Tensor, mask: Optional[torch.Tensor] = None, labels: Optional[torch.LongTensor] = None):
		assert labels is not None
		logits = self.forward(input_ids, mask)
		return F.cross_entropy(logits, labels), logits


class TextCNNClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model.classifier import TextCNNClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = TextCNNClassifier(embed_dim, len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
	             kernel_sizes=(2, 3, 4), cnn_channels: int = 64, activation=None, num_classes: int = 2,
				 num_hidden_layer: int = 0, layer_norm=False, batch_norm=False, residual=False, dropout: float = 0.0,
				 bias=False):
		"""
		:param embed_dim: word embedding维度
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param num_hidden_layer: 隐藏层数
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param residual: 是否残差
		:param dropout：
		:param bias:
		"""
		classifier = TextCNN(embed_dim, kernel_sizes, cnn_channels, num_classes, activation, num_hidden_layer,
						   layer_norm, batch_norm, residual, dropout, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class RNNAttentionClassifier(MaskEmbeddingClassifier):
	"""
	如果是英文、分词, 不用residual 效果比较好
	
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model.classifier import RNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNAttentionClassifier(embed_dim, len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, num_classes: int, embedding: Union[nn.Embedding, torch.Tensor, List] = None,
	             vocab_size: int = None, hidden_size: int = 64, num_layers: int = 1, num_heads: int = 1, rnn=nn.GRU,
	             bidirectional=True, layer_norm=False, residual=False, padding_idx: Optional[int] = None, dropout: float = 0.0):
		"""
		:param embed_dim: word embedding维度
		:param num_classes: 类别数
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param residual: 是否残差
		:param padding_idx
		:param dropout：
		"""
		classifier = RNNAttention(embed_dim, hidden_size, num_layers, num_heads, num_classes, rnn, bidirectional,
		                          layer_norm, residual, dropout)
		super().__init__(classifier, embedding, vocab_size, embed_dim, padding_idx)


class CNNRNNAttentionClassifier(EmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model.classifier import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = CNNRNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=None, num_classes: int = 2,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, bidirectional=True,
				 layer_norm=False, batch_norm=False, residual=False, dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param residual: 是否残差
		:param dropout：
		:param bias：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = CNNRNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                             num_layers, num_heads, num_classes, rnn, bidirectional, layer_norm, batch_norm,
		                             residual, dropout, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim)


class RNNCNNAttentionClassifier(MaskEmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model.classifier import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(corpus)
	>>> classifier = RNNCNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=None, num_classes: int = 2,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, bidirectional=True,
				 layer_norm=False, batch_norm=False, padding_idx: Optional[int] = None, dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param padding_idx
		:param dropout：
		:param bias：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = RNNCNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                             num_layers, num_heads, num_classes, rnn, bidirectional, layer_norm, batch_norm,
		                             dropout, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim, padding_idx)


class ResRNNCNNAttentionClassifier(MaskEmbeddingClassifier):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model.classifier import CNNRNNAttentionClassifier
	>>> tokenizer = Tokenizer(texts=corpus)
	>>> classifier = ResRNNCNNAttentionClassifier(embed_dim, num_classes=len(classes), vocab_size=tokenizer.vocab_size)
	"""

	def __init__(self, embed_dim: int, seq_length: int = 16, cnn_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=None, num_classes: int = 2,
	             embedding: Union[nn.Embedding, torch.Tensor, List] = None, vocab_size: int = None,
				 hidden_size: int = 64, num_layers: int = 1,num_heads: int = 2, rnn=nn.GRU, bidirectional=True,
				 layer_norm=False, batch_norm=False, padding_idx: Optional[int] = None, dropout: float = 0.0, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param cnn_channels: CNN out_channels
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param num_classes: 类别数
		:param embedding: 出入的embedding
		:param vocab_size: vocab size
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_heads: 抽头数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param bidirectional： 是否是双向RNN
		:param layer_norm：是否层正则化
		:param batch_norm: 是否层批则化
		:param padding_idx
		:param dropout：
		:param bias：
		"""
		cnn_channels = cnn_channels or embed_dim
		classifier = ResRNNCNNAttention(embed_dim, seq_length, cnn_channels, kernel_sizes, activation, hidden_size,
		                                num_layers, num_heads, num_classes, rnn, bidirectional, layer_norm, batch_norm,
		                                dropout, bias)
		super().__init__(classifier, embedding, vocab_size, embed_dim, padding_idx)
	

class RotaryAttentionClassifier(nn.Module):
	
	def __init__(self, embed_dim: int, max_length: int, vocab_size: Optional[int] = None, embedding=None,
	             padding_idx: Optional[int] = None, num_classes: int = 2, theta: float = 10000.0, add_cnn=False,
	             dropout: float = 0.):
		"""
		:param embed_dim: word embedding维度
		:param vocab_size: vocab size
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param padding_idx:
		:param num_classes:
		:param theta:
		:param add_cnn: 是否加入CNN
		:param dropout:
		"""
		super().__init__()
		if add_cnn:
			self.embedding = CNNEmbedding(embed_dim, vocab_size=vocab_size, embedding=embedding,
			                              seq_length=max_length // 3, padding_idx=padding_idx, dropout=dropout)
		else:
			self.embedding = EmbeddingLayer(embed_dim, vocab_size=vocab_size, embedding=embedding,
			                                padding_idx=padding_idx, dropout=dropout)
		self.attention = RotaryClassSelfAttention(embed_dim, max_length, theta)
		self.norm = RMSNorm(embed_dim)
		self.fc = nn.Linear(in_features=embed_dim, out_features=num_classes)
	
	def forward(self, input_ids: torch.Tensor, start_pos: int = 0):
		embedding = self.embedding(input_ids)
		output = self.attention(embedding, start_pos)
		output = self.norm(output)
		return self.fc(output)
		
	def fit(self, input_ids: torch.Tensor, labels: torch.LongTensor):
		logits = self.forward(input_ids)
		return F.cross_entropy(logits, labels), logits



import torch
import numpy as np
from typing import List, Union
from sklearn.model_selection import train_test_split
from torch.utils.data import TensorDataset
from ._tokenize_vec import TokenizeVec


def train_test_set(tokenize_vec: TokenizeVec, train_texts: List[str], test_texts: List[str],
                   y_train: Union[torch.LongTensor, List, np.ndarray], y_test: Union[torch.LongTensor, List, np.ndarray],
                         max_length: int = 0, n_jobs=-1):
    X_train = tokenize_vec.parallel_encode_plus(train_texts, max_length=max_length, padding='max_length',
                                        truncation=True, add_special_tokens=True,
                                        return_token_type_ids=True,return_attention_mask=True,
                                        return_tensors='pt', n_jobs=n_jobs)
    if isinstance(y_train, List) or isinstance(y_train, np.ndarray):
        y_train = torch.tensor(y_train, dtype=torch.long)

    X_test = tokenize_vec.parallel_encode_plus(test_texts, max_length=max_length, padding='max_length',
                                                truncation=True, add_special_tokens=True,
                                                return_token_type_ids=True, return_attention_mask=True,
                                                return_tensors='pt', n_jobs=n_jobs)
    if isinstance(y_test, List) or isinstance(y_test, np.ndarray):
        y_test = torch.tensor(y_test, dtype=torch.long)
    return TensorDataset(X_train, y_train), TensorDataset(X_test, y_test)


def train_test_split_set(tokenize_vec: TokenizeVec, texts: List[str], y: Union[torch.LongTensor, List, np.ndarray],
                         max_length: int = 0, test_size=0.2, n_jobs=-1, random_state=None):
    X = tokenize_vec.parallel_encode_plus(texts, max_length=max_length, padding='max_length',
                                        truncation=True, add_special_tokens=True,
                                        return_token_type_ids=True,return_attention_mask=True,
                                        return_tensors='pt', n_jobs=n_jobs)
    if isinstance(y, List) or isinstance(y, np.ndarray):
        y = torch.tensor(y, dtype=torch.long)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)
    return TensorDataset(X_train, y_train), TensorDataset(X_test, y_test)



import torch
import numpy as np
import pandas as pd
from typing import Union, List, Tuple
from torch.utils.data import Dataset


class ListDataset(Dataset[Tuple[List, ...]]):
	r"""Dataset wrapping List.

	Each sample will be retrieved by indexing list.

	Args:
		*lists (List): lists that have the same length.
	"""
	
	lists: Tuple[List, ...]
	
	def __init__(self, *lists: List) -> None:
		assert all(len(lists[0]) == len(sub_list) for sub_list in lists), "Size mismatch between tensors"
		self.lists = lists
	
	def __getitem__(self, index):
		if len(self.lists) == 1:
			return self.lists[0][index]
		return tuple(sub_list[index] for sub_list in self.lists)
	
	def __len__(self):
		return len(self.lists[0])


class TokenDataset(Dataset):
	""" Token的长度不一样
	返回的是(token, label) 不是Tensor, 必须经过PaddingTokenCollator
	"""
	
	def __init__(self, tokens: Union[List[int], np.ndarray],
	             labels: Union[List, np.ndarray, pd.Series]):
		super().__init__()
		self.tokens = tokens
		self.labels = labels.values if isinstance(labels, pd.Series) else labels
	
	def __getitem__(self, index: int):
		return self.tokens[index], self.labels[index]
	
	def __len__(self):
		return len(self.labels)


class SameLengthTokenDataset(Dataset):
	""" Token已经truncate和padding, 长度一样
	返回的是Tensor, 不需要经过collate_fn
	"""
	
	def __init__(self, tokens: Union[List[int], np.ndarray, torch.LongTensor],
	             labels: Union[List, np.ndarray, pd.Series, torch.LongTensor]):
		super().__init__()
		self.tokens = tokens if torch.is_tensor(tokens) else torch.LongTensor(tokens)
		labels = labels.values if isinstance(labels, pd.Series) else labels
		self.labels = labels if torch.is_tensor(labels) else torch.LongTensor(labels)
	
	def __getitem__(self, index: int):
		return self.tokens[index], self.labels[index]
	
	def __len__(self):
		return len(self.labels)


class TextDataset(Dataset):
	""" 返回的是(text, label) 不是Tensor, 必须经过TextVecCollator """
	
	def __init__(self, texts: Union[List[str], np.ndarray, pd.Series], labels: Union[List, np.ndarray, pd.Series]):
		super().__init__()
		self.texts = texts.values if isinstance(texts, pd.Series) else texts
		self.labels = labels.values if isinstance(labels, pd.Series) else labels
	
	def __getitem__(self, index: int):
		return self.texts[index], self.labels[index]
	
	def __len__(self):
		return len(self.labels)


class TextDFDataset(Dataset):
	""" 返回的是(text, label) 不是Tensor, 必须经过TextVecCollator """
	
	def __init__(self, data_df: pd.DataFrame):
		"""
		:param data_df: 只有两列 ['text', 'label'], 注意顺序，第一列是text, 第二列是label
		"""
		super().__init__()
		self.data = data_df.values
	
	def __getitem__(self, index: int):
		return self.data[index]
	
	def __len__(self):
		return len(self.data)



from copy import deepcopy
from typing import Tuple, Union
import torch
from torch import optim
from torch.utils.data import DataLoader, TensorDataset, Dataset
from torch.optim.lr_scheduler import LRScheduler, CosineAnnealingLR

from nlpx import log_utils
from .utils import train_epoch, train_epoch_acc, evaluate, acc_evaluate, r2_evaluate, train_epoch_r2, convert_data, \
	convert_data_acc, convert_data_r2


def get_early_stopping_rounds(epochs):
	if epochs <= 10:
		return max(2, int(0.2 * epochs))
	if epochs <= 50:
		return min(10, int(0.2 * epochs))
	return max(10, int(0.1 * epochs))


class Trainer:
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "", persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device
	             ):
		self.epochs = epochs
		self.optimizer = optimizer
		self.scheduler = scheduler
		self.lr = lr
		self.T_max = T_max
		self.batch_size = batch_size
		self.num_workers = num_workers
		self.pin_memory = pin_memory
		self.pin_memory_device = pin_memory_device
		self.persistent_workers = persistent_workers
		self.early_stopping_rounds = early_stopping_rounds or get_early_stopping_rounds(epochs)
		self.print_per_rounds = print_per_rounds
		self.checkpoint_per_rounds = checkpoint_per_rounds
		self.checkpoint_name = checkpoint_name
		self.device = device or torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
	
	def train(self, model, train_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		cnt = 0
		min_loss = float('inf')
		best_model = None
		train_losses = []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		model.train()
		for epoch in range(1, self.epochs + 1):
			avg_loss = train_epoch(model, train_loader, optimizer, self.device, scheduler, epoch, self.epochs,
			                       show_progress)
			train_losses.append(avg_loss)
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss)
			self.try_checkpoint(model, epoch)
			
			if min_loss - avg_loss > eps:
				cnt = 0
				best_model = deepcopy(model)
				min_loss = avg_loss
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(3, self.early_stopping_rounds) and cnt >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			cnt += 1
		
		return best_model, {'train_losses': train_losses}
	
	def get_optimizer_scheduler(self, model):
		scheduler = None
		if self.scheduler is not None:
			scheduler = self.scheduler
			optimizer = scheduler.optimizer
		elif self.optimizer is None:
			optimizer = optim.AdamW(model.parameters(), lr=self.lr)
		elif isinstance(self.optimizer, type):
			optimizer = self.optimizer(model.parameters(), lr=self.lr)
		else:
			optimizer = self.optimizer
		
		if scheduler is None and self.T_max and self.T_max > 0:
			scheduler = CosineAnnealingLR(optimizer, T_max=self.T_max)
		
		return optimizer, scheduler
	
	def try_checkpoint(self, model, epoch):
		if self.checkpoint_per_rounds <= 0:
			return
		
		if self.checkpoint_per_rounds == 1 or epoch % self.checkpoint_per_rounds == 0:
			torch.save(model, self.checkpoint_name)
	
	def try_print(self, do_print, show_progress, epoch, lr, loss, **kwargs):
		if self.print_per_rounds == 1 or epoch % self.print_per_rounds == 0:
			do_print(show_progress, epoch, lr, loss, **kwargs)
	
	def print(self, show_progress, epoch, lr, loss):
		if show_progress:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}\n')
		else:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}')


class SimpleTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data(X, y)
		return super().train(model, TensorDataset(X, y), collate_fn, show_progress, eps=eps)


class EvalTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0, pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False, early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
		self.eval_batch_size = eval_batch_size
		self.num_eval_workers = num_eval_workers
	
	def train(self, model, train_set: Dataset, val_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		val_loader = DataLoader(dataset=val_set, batch_size=self.eval_batch_size,
		                        num_workers=self.num_eval_workers, collate_fn=collate_fn)
		cnt = 0
		best_model = None
		min_loss = float('inf')
		train_losses, val_losses = [], []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		for epoch in range(1, self.epochs + 1):
			model.train()
			avg_loss = train_epoch(model, train_loader, optimizer, self.device, scheduler, epoch, self.epochs,
			                       show_progress)
			val_loss = evaluate(model, val_loader, self.device)
			train_losses.append(avg_loss)
			val_losses.append(val_loss)
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss,
			               val_loss=val_loss)
			self.try_checkpoint(model, epoch)
			
			if min_loss - val_loss > eps:
				cnt = 0
				best_model = deepcopy(model)
				min_loss = val_loss
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(3, self.early_stopping_rounds) and cnt >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			cnt += 1
		
		return best_model, {'train_losses': train_losses, 'val_losses': val_losses}
	
	def print(self, show_progress, epoch, lr, loss, val_loss):
		if show_progress:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, val_loss: {val_loss:.4f}\n')
		else:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, val_loss: {val_loss:.4f}')


class SimpleEvalTrainer(EvalTrainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, eval_batch_size, num_workers,
		                 num_eval_workers, pin_memory, pin_memory_device, persistent_workers, early_stopping_rounds,
		                 print_per_rounds, checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, val_data: Tuple, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data(X, y)
		X_val, y_val = convert_data(val_data[0], val_data[1])
		return super().train(model, TensorDataset(X, y), TensorDataset(X_val, y_val), collate_fn, show_progress, eps)


class ClassTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, train_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		best_model = None
		min_loss = float('inf')
		cnt, cnt2, best_acc, last_acc = 0, 0, 0.0, 0.0
		train_losses, train_accs = [], []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		model.train()
		for epoch in range(1, self.epochs + 1):
			acc, avg_loss = train_epoch_acc(model, train_loader, optimizer, self.device, scheduler, epoch,
			                                self.epochs, show_progress)
			train_accs.append(acc)
			train_losses.append(avg_loss)
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss, acc=acc)
			self.try_checkpoint(model, epoch)
			
			if acc > best_acc or (acc == best_acc and min_loss - avg_loss > eps):
				cnt, cnt2 = 0, 0
				best_acc, best_model = acc, deepcopy(model)
				last_acc, min_loss = acc, min(min_loss, avg_loss)
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			if abs(last_acc - acc) < eps:  # val_acc不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_acc = acc
			best_acc = max(best_acc, acc)
			min_loss = min(min_loss, avg_loss)
		
		return best_model, {'train_losses': train_losses, 'train_accs': train_accs}
	
	def print(self, show_progress, epoch, lr, loss, acc):
		if show_progress:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_acc: {acc:.4f}, train_loss: {loss:.4f}\n')
		else:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_acc: {acc:.4f}, train_loss: {loss:.4f}')


class SimpleClassTrainer(ClassTrainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data_acc(X, y)
		return super().train(model, TensorDataset(X, y), collate_fn, show_progress, eps=eps)


class EvalClassTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0, pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False, early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
		self.eval_batch_size = eval_batch_size
		self.num_eval_workers = num_eval_workers
	
	def train(self, model, train_set: Dataset, val_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		val_loader = DataLoader(dataset=val_set, batch_size=self.eval_batch_size,
		                        num_workers=self.num_eval_workers, collate_fn=collate_fn)
		best_model = None
		min_loss = float('inf')
		cnt, cnt2, best_acc, last_acc = 0, 0, 0.0, 0.0
		train_losses, train_accs, val_losses, val_accs = [], [], [], []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		for epoch in range(1, self.epochs + 1):
			model.train()
			acc, avg_loss = train_epoch_acc(model, train_loader, optimizer, self.device, scheduler, epoch,
			                                self.epochs, show_progress)
			val_loss, val_acc = acc_evaluate(model, val_loader, self.device)
			train_accs.append(acc)
			train_losses.append(avg_loss)
			val_accs.append(val_acc)
			val_losses.append(val_loss)
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss,
			               acc=acc, val_loss=val_loss, val_acc=val_acc)
			self.try_checkpoint(model, epoch)
			
			if val_acc > best_acc or (val_acc == best_acc and min_loss - val_loss > eps):
				cnt, cnt2 = 0, 0
				best_acc, best_model = val_acc, deepcopy(model)
				last_acc, min_loss = val_acc, min(min_loss, val_loss)
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			if abs(last_acc - val_acc) < eps:  # val_acc不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_acc = val_acc
			best_acc = max(best_acc, val_acc)
			min_loss = min(min_loss, val_loss)
		
		return best_model, {'train_losses': train_losses, 'train_accs': train_accs, 'val_losses': val_losses,
		                    'val_accs': val_accs}
	
	def print(self, show_progress, epoch, lr, loss, acc, val_loss, val_acc):
		if show_progress:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_acc: {acc:.4f}, train_loss: {loss:.4f}, '
				f'val_acc: {val_acc:.4f}, val_loss: {val_loss:.4f}\n')
		else:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_acc: {acc:.4f}, train_loss: {loss:.4f}, '
				f'val_acc: {val_acc:.4f}, val_loss: {val_loss:.4f}')


class SimpleEvalClassTrainer(EvalClassTrainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, eval_batch_size, num_workers,
		                 num_eval_workers, pin_memory, pin_memory_device, persistent_workers, early_stopping_rounds,
		                 print_per_rounds, checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, val_data: Tuple, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data_acc(X, y)
		X_val, y_val = convert_data_acc(val_data[0], val_data[1])
		return super().train(model, TensorDataset(X, y), TensorDataset(X_val, y_val), collate_fn, show_progress, eps)


class RegressTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, train_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		best_model = model
		min_loss = float('inf')
		cnt, cnt2, best_r2, last_r2 = 0, 0, -1.0, -1.0
		train_losses, train_r2s = [], []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		model.train()
		for epoch in range(1, self.epochs + 1):
			r2, avg_loss = train_epoch_r2(model, train_loader, optimizer, self.device, scheduler, epoch,
			                              self.epochs, show_progress)
			train_r2s.append(r2)
			train_losses.append(avg_loss)
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss, r2=r2)
			self.try_checkpoint(model, epoch)
			
			if r2 > best_r2 or (r2 == best_r2 and min_loss - avg_loss > eps):
				cnt, cnt2 = 0, 0
				best_r2, best_model = r2, deepcopy(model)
				last_r2, min_loss = r2, min(min_loss, avg_loss)
				continue
			
			# x次epoch的val_r2不提升或x次epoch的val_r2不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			if abs(last_r2 - r2) < eps:  # val_r2不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_r2 = r2
			best_r2 = max(best_r2, r2)
			min_loss = min(min_loss, avg_loss)
		
		return best_model, {'train_losses': train_losses, 'train_r2s': train_r2s}
	
	def print(self, show_progress, epoch, lr, loss, r2):
		if show_progress:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_r2: {r2:.4f}\n')
		else:
			log_utils.info(f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_r2: {r2:.4f}')


class SimpleRegressTrainer(RegressTrainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data_r2(X, y)
		return super().train(model, TensorDataset(X, y), collate_fn, show_progress, eps=eps)


class EvalRegressTrainer(Trainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0, pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False, early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
		self.eval_batch_size = eval_batch_size
		self.num_eval_workers = num_eval_workers
	
	def train(self, model, train_set: Dataset, val_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		val_loader = DataLoader(dataset=val_set, batch_size=self.eval_batch_size,
		                        num_workers=self.num_eval_workers, collate_fn=collate_fn)
		best_model = model
		min_loss = float('inf')
		cnt, cnt2, best_r2, last_r2 = 0, 0, -1.0, -1.0
		train_losses, train_r2s, val_losses, val_r2s = [], [], [], []
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		for epoch in range(1, self.epochs + 1):
			model.train()
			r2, avg_loss = train_epoch_r2(model, train_loader, optimizer, self.device, scheduler, epoch,
			                              self.epochs, show_progress)
			val_loss, val_r2 = r2_evaluate(model, val_loader, self.device)
			train_r2s.append(r2)
			train_losses.append(avg_loss)
			val_r2s.append(val_r2)
			val_losses.append(val_loss)
			
			self.try_print(self.print, show_progress, epoch, optimizer.param_groups[0]["lr"], avg_loss,
			               r2=r2, val_loss=val_loss, val_r2=val_r2)
			self.try_checkpoint(model, epoch)
			
			if val_r2 > best_r2 or (val_r2 == best_r2 and min_loss - val_loss > eps):
				cnt, cnt2 = 0, 0
				best_r2, best_model = val_r2, deepcopy(model)
				last_r2, min_loss = val_r2, min(min_loss, val_loss)
				continue
			
			# x次epoch的val_r2不提升或x次epoch的val_r2不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.epochs}")
				break
			
			if abs(last_r2 - val_r2) < eps:  # val_r2不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_r2 = val_r2
			best_r2 = max(best_r2, val_r2)
			min_loss = min(min_loss, val_loss)
		
		return best_model, {'train_losses': train_losses, 'train_r2s': train_r2s, 'val_losses': val_losses,
		                    'val_r2s': val_r2s}
	
	def print(self, show_progress, epoch, lr, loss, r2, val_loss, val_r2):
		if show_progress:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_r2: {r2:.4f}, '
				f'val_loss: {val_loss:.4f}, val_r2: {val_r2:.4f}\n')
		else:
			log_utils.info(
				f'Epoch-{epoch}/{self.epochs}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_r2: {r2:.4f}, '
				f'val_loss: {val_loss:.4f}, val_r2: {val_r2:.4f}')


class SimpleEvalRegressTrainer(EvalRegressTrainer):
	
	def __init__(self, epochs=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             lr=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds: int = None,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	             device=torch.device):
		super().__init__(epochs, optimizer, scheduler, lr, T_max, batch_size, eval_batch_size, num_workers,
		                 num_eval_workers, pin_memory, pin_memory_device, persistent_workers, early_stopping_rounds,
		                 print_per_rounds, checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, val_data: Tuple, collate_fn=None, show_progress=False, eps=1e-5):
		X, y = convert_data_r2(X, y)
		X_val, y_val = convert_data_r2(val_data[0], val_data[1])
		return super().train(model, TensorDataset(X, y), TensorDataset(X_val, y_val), collate_fn, show_progress, eps)
